import pandas as pd
import numpy as np

def create_synth_dataframe(n: int = 365, 
                           trend: float = 0.1, 
                           seasonality: int = 7, 
                           simulated_effect: float = 0.15, 
                           eff_n: int = 15, 
                           noise_power: float = 0.2, 
                           regressor: int = 2) -> pd.DataFrame:
    """
    Creates a synthetic dataframe with time series components.

    Args:
    - n (int): Number of periods to generate.
    - trend (float): Slope of the trend component.
    - seasonality (int): Period of the seasonality component.
    - simulated_effect (float): Multiplicative effect of the treatment.
    - eff_n (int): Number of periods the treatment lasts.
    - noise_power (float): Scale of the noise component.
    - regressor (int): Number of regressors to include in the dataframe.

    Returns:
    - df (pd.DataFrame): A pandas dataframe with columns 'ds', 'y' and the regressors created.

    Example:
    - create_synth_dataframe(n=365, trend=0.1, seasonality=7, simulated_effect=1.2, eff_n=30, noise_power=0.1, regressor=3)
    """

    # Create a time index
    control_index = pd.date_range('2022-01-01', periods=n, freq='D')
    # Create a time index
    treatment_index = pd.date_range(str(control_index.max()- pd.Timedelta(days=eff_n-1)), periods=eff_n, freq='D')

    # Create the second time series
    trend_component = np.arange(start=1, stop = n+1)* trend
    seasonality_component = np.cos(np.arange(start=1, stop = n+1) * 2 * np.pi / seasonality)
    data_control = trend_component + seasonality_component + np.random.normal(scale=noise_power, size=n)

    # Create the first time series
    data_treatment = data_control[-len(treatment_index):] * simulated_effect

    df = pd.merge(
        pd.DataFrame({'control':data_control, 'ds':control_index}),
        pd.DataFrame({'treatment':data_treatment, 'ds':treatment_index}),
        on = 'ds',
        how = 'left'
        ).fillna(0)

    df['y'] = df.control  + df.treatment

    # Create regressors
    for i in range(regressor):
        trend_component = np.arange(start=1, stop = n+1) * trend//2
        seasonality_component = np.cos(np.arange(start=1, stop = n+1) * 2 * np.pi / seasonality//int(np.random.randint(1,11)))

        regressor_data = trend_component + seasonality_component + np.random.normal(scale=noise_power, size=len(df.index))
        
        df[f'regressor{i+1}'] = regressor_data

    cols = ['ds', 'y'] + [f'regressor{i+1}' for i in range(regressor)]
    df['ds'] = pd.to_datetime(df.ds)

    # Print min and max date
    print(f"Min date: {df['ds'].min()}\nMax date: {df['ds'].max()}")

    # Print day where effect was injected
    print(f"Day where effect was injected: {df[df['treatment'] != 0]['ds'].min()}")

    #Effect injected
    print(f"Power of the effect: {simulated_effect*100}%")

    return df[cols]



import pandas as pd
import numpy as np

from scipy.stats import t
from scipy.stats import norm

from typing import Union

import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)

def min_max_scale(data, min_range, max_range):
    data_min, data_max = data.min(), data.max()
    return (data - data_min) * (max_range - min_range) / (data_max - data_min) + min_range

def structural_bootstrap(data, block_size):
    num_blocks = int(np.ceil(len(data) / block_size))
    block_indices = np.random.randint(0, len(data) - block_size + 1, size=num_blocks)
    resampled_blocks = np.array([data[i:i + block_size] for i in block_indices])
    resampled_data = resampled_blocks.flatten()[:len(data)]
    return resampled_data

def random_walk_bootstrap(bootstrap_samples, n_samples, n_steps, variable, mape):
    min_range = variable.min() * (1 - mape)
    max_range = variable.max() * (mape + 1)
    # Loop over number of bootstrap samples
    for i in range(n_samples):

      # Resample data with replacement
      bootstrap_data = np.random.choice(variable, size=len(variable))#, replace=True)
      
      # Simulate random walk based on bootstrap data
      walk = np.cumsum(np.random.randn(n_steps))
      walk -= walk[0]

      walk *= bootstrap_data.std() / walk.std()
      walk += bootstrap_data.mean()

      # Smooth the simulated random walk using a moving average filter
      smoother = 2  # the amount of smoothing on either side
      
      # Pad the beginning and end of the input array
      pad_size = smoother
      padded_walk = np.pad(walk, (pad_size, pad_size), mode='edge')
      
      # Apply the smoothing filter
      walk_smoothed = np.convolve(padded_walk, np.ones(2*smoother+1)/(2*smoother+1), mode='valid')
      walk = min_max_scale(walk_smoothed, min_range, max_range)

      bootstrap_samples[i] = walk.copy()

    return bootstrap_samples

def prior_bootstrap(bootstrap_samples, n_samples, n_steps, variable, mape):
    min_range = variable.min() * (1 - mape)
    max_range = variable.max() * (mape + 1)

    # Loop over number of bootstrap samples
    for i in range(n_samples):

      # Resample data with replacement
      bootstrap_data = np.random.choice(variable, size=len(variable))
      
      # Simulate random walk based on bootstrap data
      walk = np.cumsum(np.random.normal(loc=0, scale=bootstrap_data.std(), size=n_steps))
      walk += bootstrap_data.mean()

      walk = min_max_scale(walk, min_range, max_range)


      info = variable.values
      walk = info - (np.mean(walk) - np.mean(info))

      #Save random walk as one of the bootstrap samples
      bootstrap_samples[i] = walk.copy()
    
    return bootstrap_samples

def prob_in_distribution(data, x):
  """
    Calculate the probability of a given value being in a distribution defined by the given data.

    Args:
    - data: a list or array-like object containing the data to define the distribution
    - x: a numeric value for which to calculate the probability of being in the distribution

    Returns:
    - prob: a numeric value representing the probability of x being in the distribution

    Notes:
    - This function assumes that the data follows a normal distribution.
    - The probability is calculated as a proportion of the area under the normal curve between the minimum and
      maximum values of the data.
    - If x is outside of the range of the data, the probability is 0.0.
    - If x is exactly at the mean of the data, the probability is 0.5.
    - If x is on one side of the mean, the probability is proportional to the area of the normal curve on that side.
  """
  lower_bound, upper_bound = min(data), max(data)
  
  mean, std = np.mean(data), np.std(data)

  cdf_lower = norm.cdf(lower_bound, mean, std)
  cdf_upper = 1 - norm.cdf(upper_bound, mean, std)

  if x < lower_bound or x > upper_bound:
    return 0.0
  else:
    cdf_x = norm.cdf(x, mean, std)
    if cdf_x <= 0.5:
        return 2 * (cdf_x - cdf_lower) / (1 - cdf_lower - cdf_upper)
    else:
        return 2 * (1 - cdf_x + cdf_lower) / (1 - cdf_lower - cdf_upper)

def bootstrap_simulate(
                    variable: Union[np.array, pd.DataFrame] = None, 
                    n_samples: int = 1500, 
                    n_steps: int = None,
                    mape: float = None,
                    prio = False):
    """
    Generate an array of bootstrap samples for a given dataset.

    Args:
    - data: a numpy array or pandas dataframe containing the data to be resampled
    - n_samples: an integer representing the number of bootstrap samples to generate
    - n_steps: an integer representing the number of steps in the random walk simulation; if None, defaults to the
      length of the input data

    Returns:
    - bootstrap_samples: a numpy array of shape (n_samples, n_steps) containing the bootstrap samples

    Notes:
    - This function uses the bootstrap method to resample the input data with replacement.
    - For each bootstrap sample, a random walk is simulated based on the resampled data.
    - The random walk starts at the mean of the resampled data, and each step is a random draw from a standard normal
      distribution.
    - The variance of the random walk is adjusted to match the variance of the resampled data.
    - The resulting random walk is saved as one of the bootstrap samples.
    - The function returns an array of shape (n_samples, n_steps), where each row represents a bootstrap sample.

    Example Usage:
    >>> data = np.array([1, 2, 3, 4, 5])
    >>> bootstrap_simulate(data, n_samples=1000, n_steps=10)
    """
    # Initialize array to hold bootstrap samples
    bootstrap_samples = np.empty((n_samples, n_steps))

    if prio:
      bootstrap_samples = prior_bootstrap(bootstrap_samples, n_samples, n_steps, variable, mape)
    else:
      bootstrap_samples = random_walk_bootstrap(bootstrap_samples, n_samples, n_steps, variable, mape)
    
    return bootstrap_samples
    
def bootstrap_p_value(
                    control: Union[np.array, pd.DataFrame] = None, 
                    treatment: Union[np.array, pd.DataFrame] = None, 
                    simulations: np.array = None, 
                    alpha: float = 0.05
                    ):
    """
    Calculate the p-value for a difference in means between 
    a control group and a treatment group using the bootstrap method.

    Args:
    - control: a numpy array or pandas dataframe containing data from the control group
    - treatment: a numpy array or pandas dataframe containing data from the treatment group
    - simulations: a numpy array of bootstrap samples generated using the `bootstrap_simulate` function
    - alpha: a float representing the significance level for the hypothesis test; defaults to 0.05
    - mape: a float representing the maximum allowable percent error for the bootstrapped means; if None, no error
      adjustment is made

    Returns:
    - p_value: a list containing a single float representing the calculated p-value
    - confidence_interval: a list containing two floats representing the lower and upper bounds of the confidence
      interval for the difference in means
    - bootstrapped_means: a numpy array containing the bootstrapped means used to calculate the p-value and confidence
      interval

    Notes:
    - This function assumes that the data in both control and treatment groups are normally distributed.
    - The p-value is calculated using a two-sided hypothesis test.
    - The null hypothesis is that the means of the two groups are equal.
    - The alternative hypothesis is that the means of the two groups are not equal.
    - The p-value is calculated as the proportion of bootstrapped means that are more extreme than the observed
      difference in means between the control and treatment groups.
    - A confidence interval is also calculated for the difference in means, based on the bootstrapped means.
    - If the `mape` argument is provided, the bootstrapped means are adjusted to account for potential error in the
      bootstrap sampling process.
    - The `prob_in_distribution` function is used to calculate the p-value based on the bootstrapped means.

    Example Usage:
    >>> control_data = np.array([1, 2, 3, 4, 5])
    >>> treatment_data = np.array([2, 3, 4, 5, 6])
    >>> simulations = bootstrap_simulate(control_data, n_samples=1000, n_steps=10)
    >>> bootstrap_p_value(control_data, treatment_data, simulations, alpha=0.1, mape=0.1)
    """

    # Calculate the mean of the data
    mean = np.mean(control)
    mean_treatment = np.mean(treatment)
        
    bootstrapped_means = np.empty(len(simulations))
    
    for i in range(len(simulations)):
        bootstrapped_means[i] = simulations[i].mean()
    
    lower, upper = np.percentile(bootstrapped_means, [alpha / 2 * 100, (1 - alpha / 2) * 100])
    
    p_value = prob_in_distribution(bootstrapped_means, mean_treatment)

    return [p_value], [lower, upper], bootstrapped_means


from typing import List, Union

from pandas import DataFrame
import pandas as pd

import numpy as np
import seaborn as sns

from tscausalinference.synth_regression import synth_analysis
from tscausalinference.bootstrap import bootstrap_simulate, bootstrap_p_value
from tscausalinference.load_synth_data import create_synth_dataframe
from tscausalinference.sensitivity_regression import sensitivity_analysis
from tscausalinference.plots import plot_intervention, plot_simulations, seasonal_decompose, sensitivity_curve, plot_training, plot_diagnostic
from tscausalinference.summaries import summary, summary_intervention
from tscausalinference.evaluators import mde_area

import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)

sns.set_theme()
sns.set_context("paper")

pd.options.mode.chained_assignment = None 

class tscausalinference:
    """
    A class to perform time series causal inference using Synthetic Controls Methodology.

    Parameters:
    data (np.array or DataFrame): The time series data.
    intervention (list): The intervention period in the format ['start_date', 'end_date'].
    regressors (list, optional): A list of column names to be used as regressors in the model. Defaults to empty list.
    alpha (float, optional): The level of significance. Defaults to 0.05.
    seasonality (bool, optional): Whether or not to include seasonality in the model. Defaults to True.
    n_samples (int, optional): Number of bootstrapping samples. Defaults to 1500.
    cross_validation_steps (int, optional): Number of cross-validation steps. Defaults to 5.
    model_params (dict, optional): Additional parameters for the Prophet model. Defaults to empty dictionary.

    Returns:
        None
    """

    def __init__(self,
        data: Union[np.array, DataFrame],
        intervention: Union[List[int], List[str], List[pd.Timestamp]],
        regressors: list = [],
        alpha: float = 0.05,
        seasonality: bool = True,
        n_samples: int = 1500,
        cross_validation_steps: int = 5,
        model_params: dict = {},
        model_type = 'gam',
        autocorrelation = False
        ):

        self.data = data
        self.intervention = intervention
        self.alpha = alpha
        self.seasonality = seasonality
        self.regressors = regressors
        self.n_samples = n_samples
        self.cross_validation_steps = cross_validation_steps
        self.model_params = model_params
        self.model_type = model_type
        self.autocorrelation = autocorrelation
        
        self.string_filter = "ds >= '{}' & ds <= '{}'".format(intervention[0], intervention[1])
        
    def run(self, prior = False):
        """
        Runs the causal analysis with the specified configuration.

        Args:
            prio (bool, optional): Whether to use a priori information from seasonality and trend in the analysis. Defaults to False.

        Returns:
            self: The tscausalinference object, updated with the results of the causal analysis.      
        """
        self.data, self.pre_int_metrics, self.int_metrics, self.hyper_parameters = synth_analysis(
            df = self.data, 
            regressors = self.regressors, 
            intervention = self.intervention, 
            cross_validation_steps = self.cross_validation_steps,
            alpha = self.alpha,
            model_params = self.model_params,
            model_type = self.model_type,
            autocorrelation = self.autocorrelation
            )
    
        self.simulations = bootstrap_simulate(
            variable = self.data.query(self.string_filter).yhat, 
            n_samples = self.n_samples, 
            n_steps = len(self.data.query(self.string_filter).index),
            mape = abs(round(self.pre_int_metrics[2][1],6)) / 100,
            prio = prior
            )
        
        self.stadisticts, self.stats_ranges, self.samples_means = bootstrap_p_value(control = self.data.query(self.string_filter).yhat, 
                                                                                    treatment = self.data.query(self.string_filter).y, 
                                                                                    simulations = self.simulations
                                                                                    )

        return self
    
    def plot(self, 
              method: str = 'intervention',
              past_window: int = 5, 
              back_window: int = 25, 
              figsize: tuple = (25, 10),
              simulation_number: int = 10):
        """
        Plots the time series data.

        Parameters:
            method (str, optional): The method used to plot the data. Available options are 'intervention', 'simulations', and 'decomposition'. Defaults to 'intervention'.
            past_window (int, optional): Number of past periods to include in the plot. Defaults to 5.
            back_window (int, optional): Number of future periods to include in the plot. Defaults to 25.
            figsize (tuple, optional): The figure size. Defaults to (25, 10).
            simulation_number (int, optional): The number of simulations to include in the plot if the method is 'simulations'. Defaults to 10.

        Returns:
            None
        """
        if method not in ['intervention','simulations','decomposition']:
            error = "Your method should be defined as one of these -> ('intervention','simulations','decomposition') "
            raise TypeError(error)
        
        if method == 'intervention':
            plot_intervention(data = self.data, past_window = past_window, back_window = back_window, figsize = figsize, intervention = self.intervention)
        elif method == 'simulations':
            plot_simulations(data = self.data, past_window = past_window, back_window = back_window, figsize = figsize, simulation_number = simulation_number,
                             intervention = self.intervention, simulations = self.simulations, stadisticts = self.stadisticts, 
                             stats_ranges = self.stats_ranges, samples_means = self.samples_means)
        elif method == 'decomposition':
            seasonal_decompose(data = self.data, intervention = self.intervention, figsize = figsize)
    
    def model_parameters(self):
        return self.hyper_parameters
    
    def summarization(self, statistical_significance = 0.05, method = 'general'):
        """
        Generates a summary report for the time series causal inference analysis.

        Parameters:
            statistical_significance (float, optional): The level of significance for testing statistical significance. Defaults to 0.05.
            method (str, optional): The method used to generate the summary report. Available options are 'general' and 'detailed'. Defaults to 'general'.

        Returns:
            None
        """
        if method not in ['general','detailed']:
            error = "Your method should be defined as one of these -> ('general','detailed') "
            raise TypeError(error)

        if method == 'general':
            summary(data = self.data, statistical_significance = statistical_significance, 
            stadisticts = self.stadisticts, pre_int_metrics = self.pre_int_metrics, 
            int_metrics = self.int_metrics, intervention = self.intervention, n_samples = self.n_samples)
        elif method == 'detailed':
            summary_intervention(data = self.data, intervention = self.intervention, int_metrics = self.int_metrics) 

class synth_dataframe:
    """
    Creates a synthetic dataframe with time series components.

    Args:
    - n (int): Number of periods to generate.
    - trend (float): Slope of the trend component.
    - seasonality (int): Period of the seasonality component.
    - simulated_effect (float): Multiplicative effect of the treatment.
    - eff_n (int): Number of periods the treatment lasts.
    - noise_power (float): Scale of the noise component.
    - regressor (int): Number of regressors to include in the dataframe.

    Methods:
    - DataFrame: Returns a pandas dataframe with columns 'ds', 'y' and the regressors created.

    Example:
    >>> synth = synth_dataframe(n=365, trend=0.1, seasonality=7, simulated_effect=1.2, eff_n=30, noise_power=0.1, regressor=3)
    >>> df = synth.DataFrame()
    """

    def __init__(self,
                n: int = 365, 
                trend: float = 0.1, 
                seasonality: int = 7, 
                simulated_effect: float = 0.15, 
                eff_n: int = 15, 
                noise_power: float = 0.15, 
                regressors: int = 2):
        self.n = n
        self.seasonality = seasonality
        self.trend = trend
        self.simulated_effect = simulated_effect
        self.eff_n = eff_n
        self.noise_power = noise_power
        self.regressors = regressors

        self.df = create_synth_dataframe(n = self.n, 
                           trend = self.trend, 
                           seasonality = self.seasonality, 
                           simulated_effect = self.simulated_effect, 
                           eff_n = self.eff_n, 
                           noise_power = self.noise_power, 
                           regressor = self.regressors)
    
    def DataFrame(self):
        return self.df

class sensitivity:
    """
    The `sensitivity` class helps analyze the sensitivity of a model by performing
    a sensitivity analysis with a given dataset and model type. It supports cross-validation,
    hyperparameter optimization, and visualization of results.

    Attributes:
        df (pd.DataFrame): The input dataset.
        test_period: The period used for testing.
        cross_validation_steps (int): The number of cross-validation steps.
        alpha (float): The significance level for hypothesis testing.
        model_params (dict): Model parameters to be used in the analysis.
        regressors (list): The list of regressors to be used in the analysis.
        verbose (bool): Whether to print log messages or not.
        n_samples (int): Number of samples to be used in the analysis.
        autocorrelation (bool): Whether to consider autocorrelation in the analysis.
        model_type (str): The type of model to use in the analysis GAM or Bayesian ('gam' is the default).

    Example:
        >>> import pandas as pd
        >>> data = pd.read_csv("data.csv")
        >>> sens = sensitivity(df=data, test_period=10, alpha=0.05)
        >>> sens.run()
        >>> sens.plot(method="sensitivity")    
    """
    def __init__(self,
                df: DataFrame = pd.DataFrame(), 
                test_period = None, 
                cross_validation_steps: int = 5, 
                alpha: float = 0.05, 
                model_params: dict = {}, 
                regressors: list = [],
                verbose: bool = False,
                n_samples = 1000,
                autocorrelation = False,
                model_type='gam'):

        if not isinstance(df, pd.DataFrame):
            raise ValueError("Parameter `df` must be an instance of pandas.DataFrame")
        
        if not isinstance(test_period, list):
            raise ValueError("Parameter `test_period` must be a list")

        self.df = df
        self.test_period = test_period
        self.cross_validation_steps = cross_validation_steps
        self.alpha = alpha
        self.model_params = model_params
        self.regressors = regressors
        self.model_type = model_type
        self.autocorrelation = autocorrelation
        self.verbose = verbose
        self.n_samples = n_samples

    def run(self, prior = False):
        """
        Runs the sensitivity analysis with the specified configuration.

        Args:
            prio (bool, optional): Whether to use a priori information from seasonality and trend in the analysis. Defaults to False.

        Returns:
            self: The sensitivity object, updated with the results of the sensitivity analysis.

        Example:
            >>> sens = sensitivity(df=data, test_period=10, alpha=0.05)
            >>> sens.run()        
        """
        self.data, self.analysis, self.hyper_parameters = sensitivity_analysis(df = self.df, 
                         test_period = self.test_period, 
                         cross_validation_steps = self.cross_validation_steps, 
                         alpha = self.alpha, 
                         model_params = self.model_params, 
                         regressors= self.regressors,
                         verbose = self.verbose,
                         n_samples = self.n_samples,
                         prio = prior,
                         autocorrelation = self.autocorrelation,
                         model_type = self.model_type)
        return self
    
    def data_analysis(self):
        """
        Retrieves the results of the sensitivity analysis.

        Returns:
            pd.DataFrame: The results of the sensitivity analysis, including p-values and other statistics.

        Example:
            >>> sens = sensitivity(df=data, test_period=10, alpha=0.05)
            >>> sens.run()
            >>> analysis = sens.data_analysis()    
        """
        return self.analysis
    
    def plot(self, method = 'sensitivity', figsize=(25, 10), past_window = 25, back_window = 10):
        """
        Plots the results of the sensitivity analysis using the specified method.

        Args:
            method (str, optional): The plotting method to use ('sensitivity', 'training', or 'diagnostic'). Defaults to 'sensitivity'.
            figsize (tuple, optional): The size of the plot. Defaults to (25, 10).
            past_window (int, optional): The number of past data points to display in the training plot. Defaults to 25.
            back_window (int, optional): The number of back data points to display in the training plot. Defaults to 10.

        Raises:
            TypeError: If an invalid method is provided.

        Example:
            >>> sens = sensitivity(df=data, test_period=10, alpha=0.05)
            >>> sens.run()
            >>> sens.plot(method="sensitivity")        
        """
        
        if method not in ['sensitivity','training', 'diagnostic']:
            error = "Your method should be defined as one of these -> ('sensitivity','training', 'diagnostic')"
            raise TypeError(error)
        
        if method == 'sensitivity':
            area = mde_area(y = self.analysis.pvalue.values, x = self.analysis.index)
            return sensitivity_curve(arr1 = self.analysis.index, arr2 = self.analysis.pvalue.values, area = area, figsize = figsize)
        
        elif method == 'training':
            return plot_training(data = self.data, past_window = past_window, back_window = back_window, figsize = figsize, intervention = self.test_period)
        
        elif method == 'diagnostic':
            plot_diagnostic(data = self.data, figsize = figsize, intervention = self.test_period)
        
    
    def model_best_parameters(self):
        """
        Retrieves the best hyperparameters found during the sensitivity analysis.

        Returns:
            dict: A dictionary containing the best hyperparameters for the model.

        Example:
            >>> sens = sensitivity(df=data, test_period=10, alpha=0.05)
            >>> sens.run()
            >>> best_params = sens.model_best_parameters()
        """
        return self.hyper_parameters


import pandas as pd
import numpy as np
import datetime

def weekly_season(intervention = [], df = pd.DataFrame()):
    start_date = datetime.datetime.strptime(intervention[0], '%Y-%m-%d').date()
    end_date = datetime.datetime.strptime(intervention[1], '%Y-%m-%d').date()

    prophet_predict = df.copy()
    prophet_predict['weekday_name'] = pd.to_datetime(prophet_predict['ds']).dt.strftime("%A")

    condition = (prophet_predict.ds >= pd.to_datetime(intervention[0]) - pd.Timedelta(days=60))&(prophet_predict.ds < pd.to_datetime(intervention[0]))

    units = prophet_predict[condition].groupby('weekday_name').median().weekly
    trends = prophet_predict[condition].groupby('weekday_name').mean().trend

    day_percentages = abs(units / units.abs().sum() - 1)
    trend_percentages = abs(trends / trends.abs().sum())

    df_pctg = pd.merge(pd.DataFrame(day_percentages, index=day_percentages.index).reset_index(),
                        pd.DataFrame(trend_percentages, index=trend_percentages.index).reset_index(),
                        on='weekday_name')

    period = (end_date - start_date).days + 1

    dates = pd.date_range(intervention[0], periods = period, freq = 'D')

    df = pd.DataFrame({'date': dates})
    df['weekday_name'] = pd.to_datetime(df['date']).dt.strftime('%A')

    df = pd.merge(df, df_pctg, on='weekday_name')

    seasonality = df[(df['date'] >= pd.to_datetime(start_date)) & (df['date'] <= pd.to_datetime(end_date))].sort_values('date')['weekly'].values
    trend = df[(df['date'] >= pd.to_datetime(start_date)) & (df['date'] <= pd.to_datetime(end_date))].sort_values('date')['trend'].values
    
    return np.array(seasonality), np.array(trend)

def yearly_season(intervention = [], df = pd.DataFrame()):
    start_date = datetime.datetime.strptime(intervention[0], '%Y-%m-%d').date()
    end_date = datetime.datetime.strptime(intervention[1], '%Y-%m-%d').date()

    prophet_predict = df.copy()
    prophet_predict['year_name'] = pd.to_datetime(prophet_predict['ds']).dt.strftime("%Y")

    condition = (prophet_predict.ds >= pd.to_datetime(intervention[0]) - pd.Timedelta(days=60))&(prophet_predict.ds < pd.to_datetime(intervention[0]))

    units = prophet_predict[condition].groupby('year_name').median().yearly
    trends = prophet_predict[condition].groupby('year_name').mean().trend

    day_percentages = abs(units / units.abs().sum() - 1)
    trend_percentages = abs(trends / trends.abs().sum())

    df_pctg = pd.merge(pd.DataFrame(day_percentages, index=day_percentages.index).reset_index(),
                        pd.DataFrame(trend_percentages, index=trend_percentages.index).reset_index(),
                        on='year_name')
    
    period = (end_date - start_date).days + 1
    
    dates = pd.date_range(intervention[0], periods = period, freq = 'D')
    

    df = pd.DataFrame({'date': dates})
    df['year_name'] = pd.to_datetime(df['date']).dt.strftime('%Y')
    
    df = pd.merge(df, df_pctg, on='year_name')

    seasonality = df[(df['date'] >= pd.to_datetime(start_date)) & (df['date'] <= pd.to_datetime(end_date))].sort_values('date')['yearly'].values
    trend = df[(df['date'] >= pd.to_datetime(start_date)) & (df['date'] <= pd.to_datetime(end_date))].sort_values('date')['trend'].values
    
    return np.array(seasonality), np.array(trend)